from typing import Tuple, Optional

import torch
from omegaconf import DictConfig
from torch import nn

from commode_utils.modules import LuongAttention
from commode_utils.modules.base_decoder_step import BaseDecoderStep, DecoderState


class LSTMDecoderStep(BaseDecoderStep):
    def __init__(self, config: DictConfig, output_size: int, pad_idx: Optional[int] = None):
        super().__init__()
        self._decoder_num_layers = config.decoder_num_layers

        self._target_embedding = nn.Embedding(output_size, config.embedding_size, padding_idx=pad_idx)

        self._attention = LuongAttention(config.decoder_size)

        self._decoder_lstm = nn.LSTM(
            config.embedding_size,
            config.decoder_size,
            num_layers=config.decoder_num_layers,
            dropout=config.rnn_dropout if config.decoder_num_layers > 1 else 0,
            batch_first=True,
        )
        self._dropout_rnn = nn.Dropout(config.rnn_dropout)

        self._concat_layer = nn.Linear(config.decoder_size * 2, config.decoder_size, bias=False)
        self._norm = nn.LayerNorm(config.decoder_size)
        self._projection_layer = nn.Linear(config.decoder_size, output_size, bias=False)

    def get_initial_state(self, encoder_output: torch.Tensor, attention_mask: torch.Tensor) -> DecoderState:
        initial_state: torch.Tensor = encoder_output.sum(dim=1)  # [batch size; encoder size]
        segment_sizes: torch.Tensor = (attention_mask == 0).sum(dim=1, keepdim=True)  # [batch size; 1]
        initial_state /= segment_sizes  # [batch size; encoder size]
        initial_state = initial_state.unsqueeze(0).repeat(self._decoder_num_layers, 1, 1)
        return initial_state, initial_state

    def forward(
        self,
        input_token: torch.Tensor,
        encoder_output: torch.Tensor,
        attention_mask: torch.Tensor,
        decoder_state: DecoderState,
    ) -> Tuple[torch.Tensor, DecoderState]:
        h_prev, c_prev = decoder_state

        # [batch size; 1; embedding size]
        embedded = self._target_embedding(input_token).unsqueeze(1)

        # hidden -- [n layers; batch size; decoder size]
        # output -- [batch size; 1; decoder size]
        rnn_output, (h_prev, c_prev) = self._decoder_lstm(embedded, (h_prev, c_prev))
        rnn_output = self._dropout_rnn(rnn_output)

        # [batch size; context size]
        attn_weights = self._attention(h_prev[-1], encoder_output, attention_mask)

        # [batch size; 1; decoder size]
        context = torch.bmm(attn_weights.unsqueeze(1), encoder_output)

        # [batch size; 2 * decoder size]
        concat_input = torch.cat([rnn_output, context], dim=2).squeeze(1)

        # [batch size; decoder size]
        concat = self._concat_layer(concat_input)
        concat = self._norm(concat)
        concat = torch.tanh(concat)

        # [batch size; vocab size]
        output = self._projection_layer(concat)

        return output, (h_prev, c_prev)



import torch
from omegaconf import DictConfig
from torch import nn

from commode_utils.modules import LocalAttention
from commode_utils.training import cut_into_segments


class Classifier(nn.Module):

    _negative_value = -1e9
    _activations = {"relu": nn.ReLU, "sigmoid": nn.Sigmoid, "tanh": nn.Tanh}

    def _get_activation(self, activation_name: str) -> nn.Module:
        if activation_name in self._activations:
            return self._activations[activation_name]()
        raise KeyError(f"Activation {activation_name} is not supported")

    def __init__(self, config: DictConfig, output_size: int):
        super().__init__()

        self._attention = LocalAttention(config.classifier_size)
        hidden_layers = []
        for _ in range(config.classifier_layers):
            hidden_layers += [
                nn.Linear(config.hidden_size, config.hidden_size),
                self._get_activation(config.activation),
            ]
        hidden_layers.append(nn.Linear(config.hidden_size, output_size))
        self._layers = nn.Sequential(*hidden_layers)

    def forward(self, encoder_output: torch.Tensor, segment_sizes: torch.LongTensor) -> torch.Tensor:
        """Classify encoder output using local attention

        :param encoder_output: [n sequences; encoder size] -- stacked encoder output
        :param segment_sizes: [batch size] -- size of each segment
        :return: [batch size; logit size] -- predicted probabilities
        """
        # [batch size; max seq len; classifier size], [batch size; max seq len]
        batched_encoder_output, attention_mask = cut_into_segments(encoder_output, segment_sizes, self._negative_value)

        # [batch size; max context size; 1]
        attn_weights = self._attention(batched_encoder_output, attention_mask)

        # [batch size; classifier size]
        context = torch.bmm(attn_weights.transpose(1, 2), batched_encoder_output).squeeze(1)

        # [batch size; num classes]
        output = self._layers(context)
        output = torch.softmax(output, dim=-1)
        return output



from .attention import LuongAttention, LocalAttention
from .lstm_decoder_step import LSTMDecoderStep
from .decoder import Decoder
from .classifier import Classifier

__all__ = ["LuongAttention", "LocalAttention", "LSTMDecoderStep", "Decoder", "Classifier"]



import pickle
from abc import abstractmethod
from collections import Counter
from os.path import join, exists, dirname
from typing import Dict, Counter as CounterType, Type

from omegaconf import DictConfig
from tqdm.auto import tqdm

from commode_utils.filesystem import count_lines_in_file


class BaseVocabulary:
    vocab_filename = "vocabulary.pkl"
    _log_filename = "bad_samples.log"

    PAD = "<PAD>"
    SOS = "<SOS>"
    EOS = "<EOS>"
    UNK = "<UNK>"

    _separator = "|"
    LABEL = "label"
    TOKEN = "token"
    NODE = "node"

    def __init__(self, config: DictConfig, data_folder: str):
        vocabulary_file = join(data_folder, config.name, self.vocab_filename)
        if not exists(vocabulary_file):
            raise ValueError(f"Can't find vocabulary file ({vocabulary_file})")
        with open(vocabulary_file, "rb") as f_in:
            self._counters: Dict[str, CounterType[str]] = pickle.load(f_in)

        self._label_to_id = {self.PAD: 0, self.UNK: 1, self.SOS: 2, self.EOS: 3}
        self._label_to_id.update(
            (token[0], i + 4) for i, token in enumerate(self._counters[self.LABEL].most_common(config.max_labels))
        )

        self._token_to_id = {self.PAD: 0, self.UNK: 1, self.SOS: 2, self.EOS: 3}
        self._token_to_id.update(
            (token[0], i + 4) for i, token in enumerate(self._counters[self.TOKEN].most_common(config.max_tokens))
        )

        self._node_to_id = {self.PAD: 0, self.UNK: 1}
        self._node_to_id.update((token, i + 2) for i, token in enumerate(self._counters[self.NODE]))

    @property
    def label_to_id(self) -> Dict[str, int]:
        return self._label_to_id

    @property
    def token_to_id(self) -> Dict[str, int]:
        return self._token_to_id

    @property
    def node_to_id(self) -> Dict[str, int]:
        return self._node_to_id

    @staticmethod
    @abstractmethod
    def process_raw_sample(raw_sample: str, counters: Dict[str, CounterType[str]]):
        raise NotImplementedError()


def build_from_scratch(train_data: str, vocabulary_cls: Type[BaseVocabulary]):
    total_samples = count_lines_in_file(train_data)
    counters: Dict[str, CounterType[str]] = {
        key: Counter() for key in [vocabulary_cls.LABEL, vocabulary_cls.TOKEN, vocabulary_cls.NODE]
    }
    with open(train_data, "r") as f_in:
        for raw_sample in tqdm(f_in, total=total_samples):
            vocabulary_cls.process_raw_sample(raw_sample, counters)

    for feature, counter in counters.items():
        print(f"Count {len(counter)} {feature}, top-5: {counter.most_common(5)}")

    dataset_dir = dirname(train_data)
    vocabulary_file = join(dataset_dir, vocabulary_cls.vocab_filename)
    with open(vocabulary_file, "wb") as f_out:
        pickle.dump(counters, f_out)



from typing import Optional

from torch import nn, Tensor


class SequenceCrossEntropyLoss(nn.CrossEntropyLoss):
    """Calculate cross entropy loss with ignoring PAD index.

    :param pad_idx: an optional index of padding in target sequence.
    :param reduction: rule to reduce computed loss over batch.
    """

    __known_reductions = {"mean": lambda loss: loss.mean(), "sum": lambda loss: loss.sum()}

    def __init__(self, pad_idx: Optional[int] = None, reduction: Optional[str] = "mean"):
        super().__init__()
        self.__pad_idx = pad_idx
        if reduction is not None and reduction not in self.__known_reductions.keys():
            raise ValueError(f"Unknown reduction: {reduction}")
        self.__reduction = reduction

    def forward(self, logits: Tensor, target: Tensor) -> Tensor:
        """Calculated loss for given logits and target.

        :param logits: tensor with logits with shape [seq len; batch size; vocab size]
        :param target: tensor with target classes with shape [seq len; batch size]
        :return:
        """
        # [batch size; vocab size; seq length]
        _logits = logits.permute(1, 2, 0)
        # [batch size; seq length]
        _labels = target.permute(1, 0)
        # [batch size; seq length]
        loss = nn.functional.cross_entropy(_logits, _labels, reduction="none")
        if self.__pad_idx is not None:
            # [batch size; seq length]
            mask = _labels != self.__pad_idx
            seq_len = mask.sum(-1)
            # [batch size]
            example_loss = (loss * mask).sum(-1) / seq_len
        else:
            # [batch size]
            example_loss = loss.mean(-1)

        if self.__reduction is None:
            return example_loss
        else:
            return self.__known_reductions[self.__reduction](example_loss)



import torch
from torch import nn

from commode_utils.modules.base_decoder_step import BaseDecoderStep
from commode_utils.training import cut_into_segments


class Decoder(nn.Module):

    _negative_value = -1e9

    def __init__(self, decoder_step: BaseDecoderStep, output_size: int, sos_token: int, teacher_forcing: float = 0.0):
        super().__init__()
        self._decoder_step = decoder_step
        self._teacher_forcing = teacher_forcing
        self._out_size = output_size
        self._sos_token = sos_token

    def forward(
        self,
        encoder_output: torch.Tensor,
        segment_sizes: torch.LongTensor,
        output_size: int,
        target_sequence: torch.Tensor = None,
    ) -> torch.Tensor:
        """Generate output sequence based on encoder output

        :param encoder_output: [n sequences; encoder size] -- stacked encoder output
        :param segment_sizes: [batch size] -- size of each segment in encoder output
        :param output_size: size of output sequence
        :param target_sequence: [batch size; max seq len] -- if passed can be used for teacher forcing
        :return: [output size; batch size; vocab size] -- sequence with logits for each position
        """
        batch_size = segment_sizes.shape[0]
        batched_encoder_output, attention_mask = cut_into_segments(encoder_output, segment_sizes, self._negative_value)

        decoder_state = self._decoder_step.get_initial_state(batched_encoder_output, attention_mask)

        # [output size; batch size; vocab size]
        output = batched_encoder_output.new_zeros((output_size, batch_size, self._out_size))
        # [batch size]
        current_input = batched_encoder_output.new_full((batch_size,), self._sos_token, dtype=torch.long)
        for step in range(output_size):
            current_output, decoder_state = self._decoder_step(
                current_input, batched_encoder_output, attention_mask, decoder_state
            )
            output[step] = current_output
            if self.training and target_sequence is not None and torch.rand(1) <= self._teacher_forcing:
                current_input = target_sequence[step]
            else:
                current_input = output[step].argmax(dim=-1)

        return output



from abc import abstractmethod
from typing import Union, Tuple

import torch
from torch import nn

DecoderState = Union[torch.Tensor, Tuple[torch.Tensor, ...]]


class BaseDecoderStep(nn.Module):
    @abstractmethod
    def get_initial_state(self, encoder_output: torch.Tensor, attention_mask: torch.Tensor) -> DecoderState:
        """Use this function to initialize decoder state, e.g. h and c for LSTM decoder.

        :param encoder_output: [batch size; max seq len; encoder size] -- encoder output
        :param attention_mask: [batch size; max seq len] -- mask with zeros on non pad elements
        :return: decoder state: single tensor of tuple of tensors
        """
        raise NotImplementedError()

    @abstractmethod
    def forward(
        self,
        input_token: torch.Tensor,
        encoder_output: torch.Tensor,
        attention_mask: torch.Tensor,
        decoder_state: DecoderState,
    ) -> Tuple[torch.Tensor, DecoderState]:
        """Perform one decoder step based on input token and encoder output.

        :param input_token: [batch size] -- tokens from previous step or from target sequence
        :param encoder_output: [batch size; max seq len; encoder size] -- encoder output
        :param attention_mask: [batch size; max seq len] -- mask with zeros on non pad elements
        :param decoder_state: decoder state from previous or initial step
        :return: [batch size; vocab size] -- logits and new decoder step
        """
        raise NotImplementedError()



from os.path import split, join
from typing import Dict, List, Optional, Union

import torch
from pytorch_lightning import Callback, Trainer, LightningModule
from pytorch_lightning.loggers import WandbLogger

from commode_utils.common import print_table


class UploadCheckpointCallback(Callback):
    """Upload checkpoints after every validation epoch. Works only for wandb."""

    def __init__(self, checkpoint_dir: str):
        super().__init__()
        self._checkpoint_dir = checkpoint_dir

    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule):
        logger = trainer.logger
        if isinstance(logger, WandbLogger):
            experiment = logger.experiment
            root_dir, _ = split(self._checkpoint_dir)
            experiment.save(join(self._checkpoint_dir, "*.ckpt"), base_path=root_dir)


class PrintEpochResultCallback(Callback):
    """Print metrics to stdout after each specified epoch."""

    def __init__(
        self, after_train: bool = True, after_validation: bool = True, after_test: bool = True, split_symbol: str = "/"
    ):
        super().__init__()
        self.__after_train = after_train
        self.__after_validation = after_validation
        self.__after_test = after_test
        self.__split_symbol = split_symbol

    def _get_values(self, logged_metrics: Dict[str, Union[float, torch.Tensor]], step: str) -> List:
        values = []
        for key, value in logged_metrics.items():
            if self.__split_symbol not in key:
                continue
            group, metric = key.split(self.__split_symbol, 1)
            if group not in step:
                continue
            if isinstance(value, torch.Tensor):
                value = value.item()
            values.append(f"{metric} = {round(value, 2)}")
        return values

    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule, unused: Optional[bool] = None):
        if not self.__after_train:
            return
        metric = {"train": self._get_values(trainer.logged_metrics, "train")}
        if f"val{self.__split_symbol}loss" in trainer.logged_metrics:
            metric["val"] = self._get_values(trainer.logged_metrics, "val")
        print_table(metric)

    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule):
        if not self.__after_validation:
            return
        metric = {"val": self._get_values(trainer.logged_metrics, "val")}
        print_table(metric)

    def on_test_epoch_end(self, trainer: Trainer, pl_module: LightningModule):
        if not self.__after_test:
            return
        metric = {"test": self._get_values(trainer.logged_metrics, "test")}
        print_table(metric)



from setuptools import setup

VERSION = "0.3.4"

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("requirements.txt") as requirements_file:
    install_requires = [line.strip() for line in requirements_file]

setup_args = dict(
    name="commode-utils",
    version=VERSION,
    description="Set of useful functions and modules for Code Modeling",
    long_description_content_type="text/markdown",
    long_description=readme,
    install_requires=install_requires,
    license="Apache 2.0",
    package_data={"commode_utils": ["py.typed"]},
    packages=["commode_utils", "commode_utils.metrics", "commode_utils.modules", "commode_utils.losses"],
    zip_safe=False,
    author="Egor Spirin",
    author_email="spirin.egor@gmail.com",
    keywords=[],
    url="https://github.com/SpirinEgor/commode-utils",
    download_url="https://pypi.org/project/commode-utils/",
)

if __name__ == "__main__":
    setup(**setup_args)
